"""
Node Offset Tool by Taza

Information: This tool is used to change node coordinates in .jbeam files.
Jbeam is a close relative to .json, and can be opened with any text editor.
Nodes are defined as a list in jbeam (example file lines 13-28, 111-158).
Jbeam files are a definition of structure for the PC game "BeamNG.drive".

Usage:
1. Launch the application, and input the path to your .jbeam file in the
"Jbeam file path:" entry, or use the "Browse" button to get the file path to
your file (hatch_wheeliebar.jbeam is included as an example file).
2. Enter the name of your part in the "Part name:" entry
("node_offset_test", or "node_offset_test_1" in example file) without quotes.
3. Write the offset you want in the corresponding entries. Any axis left empty
will be interpreted as 0.
4. Optionally, you can tick the "Save changes" checkbox. This will write the
old and new node locations to a text file in the same directory as your jbeam
file.
5. Click "Apply Offset", and the offset will be applied to your file. Note:
this writes over the old file.
6. Your offset is now applied, and any information or errors will be printed
on the bottom of the application. Changes will be visible once you reload the
file in notepad++ or relaunch notepad, or press Ctrl+R in-game.

For support and updates, see:
https://www.beamng.com/threads/node-offset-tool.69831/

Made by Tapio Kauppinen aka Taza
"""
import tkinter
from jbeam import Jbeam
from tkinter import *
from tkinter import ttk
from tkinter import filedialog
import os
import time


class NodeOffsetUi:
    def __init__(self):

        self.__background = "#aaffaa"
        self.__filename = ""
        self.__baseHeight = 196
        self.__version = "0.3"

        # main window for node offset tool
        self.__offsetWindow = Tk()
        self.__offsetWindow.title("Node Offset Tool v" + self.__version)
        self.__offsetWindow.minsize(340, self.__baseHeight)
        self.__offsetWindow.maxsize(340, self.__baseHeight)
        self.__offsetWindow.configure(background=self.__background)
        self.__offsetWindow.bind('<Escape>', self.quit)
        try:
            self.__offsetWindow.iconbitmap("beamng.ico")
        except TclError:
            print("Icon file missing...")

        # static top text
        self.__topText = Label(self.__offsetWindow,
                               text="I'm not responsible for anything.",
                               background=self.__background,
                               foreground="#ff55ff",
                               font=("TkDefaultFont", 7))
        self.__topText1 = Label(self.__offsetWindow,
                                text="Created by Taza.",
                                background=self.__background,
                                foreground="#ff55ff",
                                font=("TkDefaultFont", 7))

        # explanation texts for inputs
        self.__pathText = Label(self.__offsetWindow, text="Jbeam file path:",
                                background=self.__background)
        self.__partText = Label(self.__offsetWindow, text="Part name:",
                                background=self.__background)
        self.__offsetText = Label(self.__offsetWindow, text="Offset:",
                                  background=self.__background)
        self.__XText = Label(self.__offsetWindow, text="X:",
                             background=self.__background)
        self.__YText = Label(self.__offsetWindow, text="Y:",
                             background=self.__background)
        self.__ZText = Label(self.__offsetWindow, text="Z:",
                             background=self.__background)

        # bottom text for errors and success
        self.__outputText = Label(self.__offsetWindow,
                                  text="", justify=LEFT,
                                  background=self.__background,
                                  font="TkFixedFont")

        # input fields
        self.__pathField = Entry(width=28)
        self.__partField = Entry()
        self.__offsetXField = Entry(width=6)
        self.__offsetYField = Entry(width=6)
        self.__offsetZField = Entry(width=6)

        # buttons
        self.__offsetButton = Button(self.__offsetWindow, text="Apply Offset",
                                     command=self.offset, width=15)
        self.__quitButton = Button(self.__offsetWindow, text="Quit (ESC)",
                                   command=self.quit, width=15)
        self.__browseButton = Button(self.__offsetWindow, text="Browse",
                                     command=self.file_dialog)

        # checkbox
        self.__save = IntVar()
        self.__saveToText = Checkbutton(self.__offsetWindow,
                                        variable=self.__save,
                                        text="Save changes to a text file",
                                        background=self.__background)

        # separator
        self.__separator = tkinter.ttk.Separator(self.__offsetWindow,
                                                 orient=HORIZONTAL)

        # text placement
        self.__topText.grid(row=0, column=0, columnspan=6, sticky=W)
        self.__topText1.grid(row=0, column=7, columnspan=2, sticky=E)
        self.__pathText.grid(row=1, column=0, columnspan=6, sticky=W)
        self.__partText.grid(row=3, column=0, columnspan=6, sticky=W)
        self.__offsetText.grid(row=5, column=0, columnspan=6, sticky=W)
        self.__XText.grid(row=6, column=1, sticky=E)
        self.__YText.grid(row=6, column=3, sticky=E)
        self.__ZText.grid(row=6, column=5, sticky=E)
        self.__outputText.grid(row=9, column=0, columnspan=8, sticky=W+N)

        # input field placement
        self.__pathField.grid(row=2, column=0, columnspan=6,
                              sticky=E+W, padx=2)
        self.__partField.grid(row=4, column=0, columnspan=6,
                              sticky=E+W, padx=2)
        self.__offsetXField.grid(row=6, column=2, sticky=W)
        self.__offsetYField.grid(row=6, column=4, sticky=W)
        self.__offsetZField.grid(row=6, column=6, sticky=W)

        # button placement
        self.__browseButton.grid(row=2, column=6)
        self.__offsetButton.grid(row=2, column=7, rowspan=3, sticky=N+S)
        self.__quitButton.grid(row=5, column=7, rowspan=3, sticky=N+S)

        # checkbox placement
        self.__saveToText.grid(row=7, column=1, columnspan=6)

        # separator placement
        self.__separator.grid(row=8, column=0, columnspan=8, sticky=E+W)

    def offset(self):
        """
        Offset all nodes based on values written on ui
        :return: None
        """
        # check inputs
        self.__filename = self.__pathField.get()
        if not self.__filename:
            self.update_message("No file specified!")
            self.__pathField.focus()
            return
        if not self.__filename.endswith(".jbeam"):
            self.update_message("File format not .jbeam!")
            self.__pathField.focus()
            return
        part_name = self.__partField.get()
        if not part_name:
            self.update_message("No part name specified!")
            self.__partField.focus()
            return

        x = self.__offsetXField.get()
        if x != "":
            try:
                x = float(x)
            except ValueError:
                self.update_message("Invalid offset on X axis!")
                self.__offsetXField.focus()
                return
        else:
            x = 0

        y = self.__offsetYField.get()
        if y != "":
            try:
                y = float(y)
            except ValueError:
                self.update_message("Invalid offset on Y axis!")
                self.__offsetYField.focus()
                return
        else:
            y = 0

        z = self.__offsetZField.get()
        if z != "":
            try:
                z = float(z)
            except ValueError:
                self.update_message("Invalid offset on Z axis!")
                self.__offsetZField.focus()
                return
        else:
            z = 0

        if x == y == z == 0:
            self.update_message("No offset defined!")
            return

        # new jbeam file
        file = Jbeam(self.__filename)

        # load nodes to object
        try:
            if file.load_nodes(part_name) == 1:
                self.update_message("Part " + part_name + " not found!")
                self.__partField.focus()
                return
        except FileNotFoundError:
            self.update_message("File " + self.__filename + " not found!")
            self.__pathField.focus()
            return

        # offset nodes
        try:
            changes = file.offset_xyz(x, y, z)
        except (ValueError, IndexError):
            self.update_message("Error occurred while calculating new node "
                                "coordinates! Is the file in correct jbeam "
                                "format?")
            return

        success_message = "Node offset successful! "

        # check if user wants changes to text file
        if self.__save.get() == 1:
            try:
                directory = os.path.dirname(self.__filename)
                t = time.localtime()
                current_time = time.strftime("%H_%M_%S", t)
                filename = os.path.join(directory, current_time + "_" +
                                        part_name + ".txt")
                changes_file = open(filename, "x")
                for line in changes:
                    changes_file.write(line)
                    changes_file.write("\n")
                changes_file.close()
                success_message += "Text file created successfully!"
            except FileNotFoundError:
                success_message += "Failed to create text file!"

        # write new nodes back to file
        file.write_nodes()
        self.update_message(success_message)

    def start(self):
        """ Starts the mainloop.
        """
        self.__offsetWindow.mainloop()

    def quit(self, *args):
        """Quits the program.
        """
        self.__offsetWindow.destroy()

    def file_dialog(self):
        """
        Opens the unpacked folder, gets filename to self, updates entry
        :return: None
        """
        unpacked = os.path.expanduser("~/Documents/Beamng.drive/mods/"
                                      "unpacked/")
        if not os.path.exists(unpacked):
            if not os.path.exists(os.path.expanduser("~/Documents/")):
                self.update_message("Documents directory not available! "
                                    "Defaulting to '/'")
                unpacked = "/"
            else:
                self.update_message(unpacked + " is not available! "
                                               "Defaulting to '/Documents/'")
                unpacked = os.path.expanduser("~/Documents/")
        filename = filedialog.askopenfilename(initialdir=unpacked,
                                              title="Select Jbeam file",
                                              filetype=(("Jbeam", "*.jbeam"),
                                                        ("All files", "*.*")))
        # do not act if user pressed cancel
        if not filename:
            return
        self.__filename = filename
        self.__pathField.delete(0, END)
        self.__pathField.insert(0, self.__filename)

    def update_message(self, message):
        """
        Updates the message at the bottom of the offset tool
        :param message: text to be displayed on bottom
        :return: None
        """
        # cut message to fit window
        t = time.localtime()
        current_time = time.strftime("%H:%M:%S", t)
        message = current_time + " " + message
        lines = 0
        words = message.split(" ")
        message = [""]
        WIDTH = 42
        for word in words:
            # single word, such as file location, is longer than width
            if len(word) > WIDTH:
                while len(word) > WIDTH:
                    message.append(word[0:WIDTH])
                    lines += 1
                    word = word[WIDTH:len(word)]
                message.append(word + " ")
                lines += 1

            # line going over width limit
            elif len(message[lines]) + len(word) > WIDTH:
                lines += 1
                message.append(word + " ")

            # word fits in line
            else:
                message[lines] += word + " "

        # update message and window size
        self.__outputText.configure(text="\n".join(list(map(lambda a: a.strip()
                                                            , message))))
        self.__offsetWindow.minsize(340, self.__baseHeight+16*lines)
        self.__offsetWindow.maxsize(340, self.__baseHeight+16*lines)


def main():
    ui = NodeOffsetUi()
    ui.start()


main()
